/* posix_fallocate() and posix_fallocate64() wrapper to use new 
 * fallocate() system call
 *
 * Compile this as a library and put it in LD_PRELOAD for programs
 * that use posix_fallocate() where you want them to use
 * fallocate() instead. 
 *
 * Optionally, if WRAP_FTRUNCATE is set, ftruncate() and 
 * ftruncate64() are wrapped as well. 
 * 
 * Requires kernel 2.6.23 or above for fallocate() call, and
 * a filesystem that implements it (e.g. ext4).
 *
 * compile:
 * gcc -O2 -fPIC -shared -ldl -D_GNU_SOURCE -o fallocate.so fallocate.c
 *
 * use:
 * LD_PRELOAD=/path/to/fallocate.so <program>
 *
 * This program has not been extensively tested. Please report
 * bugs or suggestions to dev@jaysonking.com.
 */

/* Author: Jayson R. King <dev@jaysonking.com>
 * Jan 2008
 */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include<dlfcn.h>
#include<errno.h>
#include<sys/syscall.h>
#include<sys/types.h>

int posix_fallocate(int fd, __off_t offset, __off_t len)
{
	static int (*fallback_func)(int, __off_t, __off_t);
	/* do the syscall */
	int res = (int)syscall(__NR_fallocate, fd, 0, (loff_t)offset, (loff_t)len);
	if (res == -1 && (errno == EOPNOTSUPP || errno == ENOSYS))
	{
		/* not supported, fall back */
		if (!fallback_func)
			fallback_func = (int (*)(int, __off_t, __off_t))
				dlsym(RTLD_NEXT, "posix_fallocate");
		return fallback_func(fd, offset, len);
	}
	return res;
}
#ifdef WRAP_FTRUNCATE
int ftruncate(int fd, __off_t len)
{
	return posix_fallocate(fd, 0, len);
}
#endif

#ifdef __USE_LARGEFILE64
int posix_fallocate64(int fd, __off64_t offset, __off64_t len)
{
	static int (*fallback_func)(int, __off64_t, __off64_t);
	/* do the syscall */
	int res = (int)syscall(__NR_fallocate, fd, 0, (loff_t)offset, (loff_t)len);
	if (res == -1 && (errno == EOPNOTSUPP || errno == ENOSYS))
	{
		/* not supported, fall back */
		if (!fallback_func)
			fallback_func = (int (*)(int, __off64_t, __off64_t))
				dlsym(RTLD_NEXT, "posix_fallocate64");
		return fallback_func(fd, offset, len);
	}
	return res;
}
#ifdef WRAP_FTRUNCATE
int ftruncate64(int fd, __off64_t len)
{
	return posix_fallocate64(fd, 0, len);
}
#endif
#endif
